import ipywidgets as widgets
from traitlets import Unicode, Dict

@widgets.register
class ReactJupyterWidget(widgets.DOMWidget):
    """An example widget."""
    _view_name = Unicode('ReactView').tag(sync=True)
    _model_name = Unicode('ReactModel').tag(sync=True)
    _view_module = Unicode('hnx-widget').tag(sync=True)
    _model_module = Unicode('hnx-widget').tag(sync=True)
    _view_module_version = Unicode('^0.1.0').tag(sync=True)
    _model_module_version = Unicode('^0.1.0').tag(sync=True)

    component = Unicode().tag(sync=True)
    props = Dict().tag(sync=True)

    def __init__(self, **kwargs):
        super().__init__()

        self.component = self.__class__.__name__
        self.props = kwargs




# Copyright © 2018 Battelle Memorial Institute
# All rights reserved.

from itertools import combinations

import numpy as np

import networkx as nx

def inflate(items, v):
    if type(v) in {str, tuple, int, float}:
        return [v]*len(items)
    elif callable(v):
        return [v(i) for i in items]
    elif type(v) not in {list, np.ndarray} and hasattr(v, '__getitem__'):
        return [v[i] for i in items]
    return v

def inflate_kwargs(items, kwargs):
    '''
    Helper function to expand keyword arguments.

    Parameters
    ----------
    n: int
        length of resulting list if argument is expanded
    kwargs: dict
        keyword arguments to be expanded

    Returns
    -------
    dict
        dictionary with same keys as kwargs and whose values are lists of length n
    '''

    return {
        k: inflate(items, v)
        for k, v in kwargs.items()
    }

def transpose_inflated_kwargs(inflated):
    return [
        dict(zip(inflated, v))
        for v in zip(*inflated.values())
    ]

def get_frozenset_label(S, count=False, override={}):
    '''
    Helper function for rendering the labels of possibly collapsed nodes and edges

    Parameters
    ----------
    S: iterable
        list of entities to be labeled
    count: bool
        True if labels should be counts of entities instead of list

    Returns
    -------
    dict
        mapping of entity to its string representation
    '''
    def helper(v):
        if type(v) == frozenset:
            if count and len(v) > 1:
                return f'x {len(v)}'
            elif count:
                return ''
            else:
                return ', '.join([str(override.get(s, s)) for s in v])
        return str(v)

    return {v: override.get(v, helper(v)) for v in S}

def get_line_graph(H, collapse=True):
    '''
    Computes the line graph, a directed graph, where a directed edge (u, v)
    exists if the edge u is a subset of the edge v in the hypergraph.

    Parameters
    ----------
    H: Hypergraph
        the entity to be drawn 
    collapse: bool
        True if edges should be added if hyper edges are identical

    Returns
    -------
    networkx.DiGraph
        A directed graph
    '''
    D = nx.DiGraph()

    V = {edge: set(nodes)
         for edge, nodes in H.edges.elements.items()}
    
    D.add_nodes_from(V)

    for u, v in combinations(V, 2):
        if V[u] != V[v] or not collapse:
            if V[u].issubset(V[v]):
                D.add_edge(u, v)
            elif V[v].issubset(V[u]):
                D.add_edge(v, u)

    return D

def get_set_layering(H, collapse=True):
    '''
    Computes a layering of the edges in the hyper graph.

    In this layering, each edge is assigned a level. An edge u will be above
    (e.g., have a smaller level value) another edge v if v is a subset of u.

    Parameters
    ----------
    H: Hypergraph
        the entity to be drawn 
    collapse: bool
        True if edges should be added if hyper edges are identical

    Returns
    -------
    dict
        a mapping of vertices in H to integer levels
    '''

    D = get_line_graph(H, collapse=collapse)

    levels = {}

    for v in nx.topological_sort(D):
        parent_levels = [levels[u] for u, _ in D.in_edges(v)]
        levels[v] = max(parent_levels) + 1 if len(parent_levels) else 0

    return levels



import ipywidgets as widgets
from traitlets import Unicode

# See js/lib/example.js for the frontend counterpart to this file.

@widgets.register
class HelloWorld(widgets.DOMWidget):
    """An example widget."""

    # Name of the widget view class in front-end
    _view_name = Unicode('HelloView').tag(sync=True)

    # Name of the widget model class in front-end
    _model_name = Unicode('HelloModel').tag(sync=True)

    # Name of the front-end module containing widget view
    _view_module = Unicode('hnx-widget').tag(sync=True)

    # Name of the front-end module containing widget model
    _model_module = Unicode('hnx-widget').tag(sync=True)

    # Version of the front-end module containing widget view
    _view_module_version = Unicode('^0.1.0').tag(sync=True)
    # Version of the front-end module containing widget model
    _model_module_version = Unicode('^0.1.0').tag(sync=True)

    # Widget specific property.
    # Widget properties are defined as traitlets. Any property tagged with `sync=True`
    # is automatically synced to the frontend *any* time it changes in Python.
    # It is synced back to Python from the frontend *any* time the model is touched.
    value = Unicode('Hello World!').tag(sync=True)



from .react_jupyter_widget import ReactJupyterWidget

import ipywidgets as widgets
from traitlets import Dict

import pandas as pd

import matplotlib.pyplot as plt
from matplotlib.colors import to_rgba_array, to_hex
import numpy as np

from .util import get_set_layering, inflate_kwargs

from itertools import chain

converters = {
    'edgecolor': 'Stroke',
    'edgecolors': 'Stroke',
    'facecolor': 'Fill',
    'facecolors': 'Fill',
    'color': 'Fill',
    'colors': 'Fill',
    'linewidths': 'StrokeWidth',
    'linewidth': 'StrokeWidth',
}

def to_camel_case(s):
    return ''.join([
        si.title() if i > 0 else si
        for i, si in enumerate(s.split('_'))
    ])

def prepare_kwargs(items, kwargs, prefix=''):
    return {
        prefix + converters.get(k, k):
            dict(zip(items, hex_array(v) if 'color' in k else v))
        for k, v in inflate_kwargs(items, kwargs).items()
    }

def rename_kwargs(**kwargs):
    return {
        converters.get(k, to_camel_case(k)): v
        for k, v in kwargs.items()
    }

def hex_array(values):
    return [
        to_hex(c, keep_alpha=True)
        for c in to_rgba_array(values)
    ]

def hnx_kwargs_to_props(V, E,
    nodes_kwargs={},
    edges_kwargs={},
    node_labels_kwargs={},
    edge_labels_kwargs={},
    **kwargs
):
    # reproduce default hnx coloring behaviors
    edges_kwargs = edges_kwargs.copy()
    edges_kwargs.setdefault('edgecolors', plt.cm.tab10(np.arange(len(E))%10))
    edges_kwargs.setdefault('linewidths', 2)
    
    # props = kwargs.copy()
    props = {}
    props.update(prepare_kwargs(V, nodes_kwargs, prefix='node'))
    props.update(prepare_kwargs(V, node_labels_kwargs, prefix='nodeLabel'))
    props.update(prepare_kwargs(E, edges_kwargs, prefix='edge'))
    props.update(prepare_kwargs(E, edge_labels_kwargs, prefix='edgeLabel'))
    
    # if not otherwise specified, set the edge label color
    # to be the same as the edge color
    props.setdefault('edgeLabelColor', props['edgeStroke'])

    return {**props, **rename_kwargs(**kwargs)}


def to_dict_without_nans(df):
    df = df.copy()
    df.index = df.index.astype(str)

    return {
        c: df[c].dropna().to_dict()
        for c in df
    }

def incidence_dict_astype_str(H):
    return {
        str(k): list(map(str, v))
        for k, v in H.items()
    }

@widgets.register
class HypernetxWidgetView(ReactJupyterWidget):
    pos = Dict().tag(sync=True)
    node_fill = Dict().tag(sync=True)
    edge_stroke = Dict().tag(sync=True)
    selected_nodes = Dict().tag(sync=True)
    selected_edges = Dict().tag(sync=True)
    hidden_nodes = Dict().tag(sync=True)
    hidden_edges = Dict().tag(sync=True)
    removed_nodes = Dict().tag(sync=True)
    removed_edges = Dict().tag(sync=True)

    @property
    def state(self):
        return {
            'pos': self.pos,
            'node_fill': self.node_fill,
            'edge_stroke': self.edge_stroke,
            'selected_nodes': self.selected_nodes,
            'selected_edges': self.selected_edges,
            'hidden_nodes': self.hidden_nodes,
            'hidden_edges': self.hidden_edges,
            'removed_nodes': self.removed_nodes,
            'removed_edges': self.removed_edges
        }

    def get_index(self, selection={}):
        return [
            k
            for k, v in selection.items()
            if v
        ]

    @property
    def selected_node_data(self):
        return self.node_data.loc[self.get_index(self.selected_nodes)]

    @property
    def selected_edge_data(self):
        return self.edge_data.loc[self.get_index(self.selected_edges)]

    def __init__(self, H,
        collapse=True,
        node_styles={},
        with_color=True,
        ignore_hnx_properties=False,
        **kwargs
    ):
        incidence_dict = H.edges.incidence_dict\
            if H.__class__.__name__ == 'Hypergraph'\
            else H

        incidence_dict = incidence_dict_astype_str(incidence_dict)

        def get_property(id, value, default):
            if value is None:
                return default
            elif hasattr(value, 'get'):
                return value.get(id, default)
            else:
                return value
            
        V = set(chain.from_iterable(incidence_dict.values()))
        E = list(incidence_dict)

        nodes = [{'uid': uid} for uid in V]
        
        # js friendly representation of the hypergraph
        edges = [
            {
                'uid': uid,
                'elements': elements
            }
            for uid, elements in incidence_dict.items()
        ]

        # used to convert Hnx 1.2 style properties into dataframes compatible with widget
        def prepare_properties(props):
            df = pd.concat([
                props.drop(columns=['properties', 'uid']),
                pd.DataFrame(props.properties.tolist(), index=props.properties.index)
            ], axis=1)
            
            col_mask = (~df.isnull()).any(axis=0)
            
            return df[df.columns[col_mask]]

        if not ignore_hnx_properties and hasattr(H, 'edges') and hasattr(H, 'nodes'):
            # check if Hypergraph object contains properties
            kwargs['edge_data'] = prepare_properties(H.edges.properties.loc[0])
            kwargs['node_data'] = prepare_properties(H.nodes.properties.loc[0])

        if 'node_data' in kwargs:
            self.node_data = kwargs['node_data']
            kwargs['node_data'] = to_dict_without_nans(self.node_data.T)

        if 'edge_data' in kwargs:
            self.edge_data = kwargs['edge_data']
            kwargs['edge_data'] = to_dict_without_nans(self.edge_data.T)
            
        super().__init__(
            nodes=nodes,
            edges=edges,
            **hnx_kwargs_to_props(V, E, **kwargs)
        )

@widgets.register
class HypernetxWidget(HypernetxWidgetView):
    pass



import pandas as pd

import hypernetx as hnx
from hypernetx import algorithms as algo

def compute_stats(H):
    
    def compute_stats_helper(H, stats):
        return pd.DataFrame({
            k: func(H)
            for k, func in stats.items()
        })

    node_stats = {
        'Degree': lambda H: dict(zip(H, hnx.degree_dist(H)))
    }

    edges_stats = {
        'Centrality': algo.s_centrality_measures.s_betweenness_centrality
    }

    D = H.dual()
    
    node_data = pd.concat((compute_stats_helper(H, node_stats), compute_stats_helper(D, edges_stats)), axis=1)
    edge_data = pd.concat((compute_stats_helper(D, node_stats), compute_stats_helper(H, edges_stats)), axis=1)
    
    return {
        'node_data': node_data,
        'edge_data': edge_data
    }